exception Error of string

let rec replace e id e' = match e with
  | Expr.Int x                           -> Expr.Int x
  | Expr.Bool b                          -> Expr.Bool b
  | Expr.Var id'           when id' = id -> e'
  | Expr.Var id'                         -> Expr.Var id'
  | Expr.Op (op,l)                       -> Expr.Op (op, List.map (fun f -> replace f id e') l)
  | Expr.Let (id',f,f')    when id' = id -> Expr.Let (id, replace f id e', f')
  | Expr.Let (id',f,f')                  -> Expr.Let (id', replace f id e', replace f' id e')
  | Expr.LetRec (id',f,f') when id' = id -> Expr.LetRec (id, f, f')
  | Expr.LetRec (id',f,f')               -> Expr.LetRec (id', replace f id e', replace f' id e')
  | Expr.Fun (id',f)       when id' = id -> Expr.Fun (id, f)
  | Expr.Fun (id',f)                     -> Expr.Fun (id', replace f id e')
  | Expr.App (f,f')                      -> Expr.App (replace f id e', replace f' id e')

let eval e =
  let rec eval' e env = match e with
    | Expr.Int x            -> Val.Int x
    | Expr.Bool b           -> Val.Bool b
    | Expr.Var id           -> (try Env.find env id
				with Not_found -> raise (Error ("Variable '"^id^"' is unbound in current environment.")))
    | Expr.Op (op,l)        -> eval_op (op,l) env
    | Expr.Let (id,f,f')    -> let vf = eval' f env in eval' f' (Env.add env id vf)
    | Expr.LetRec (id,f,f') -> let vf = eval' (replace f id (Expr.LetRec (id,f,f))) env in eval' f' (Env.add env id vf)
    | Expr.Fun (id,f)       -> Val.Fun (id, f, Env.copy env)
    | Expr.App (f,f')       -> (match eval' f env with
				  | Val.Fun (id,f,env') -> let vf' = eval' f' env in eval' f (Env.add env' id vf')
				  | _ -> raise (Error "Type checking error in App."))

  and eval_op (op,l) env = match (op,l) with
    | (Expr.Add, [x;y])  -> (match (eval' x env, eval' y env) with
			       | (Val.Int vx, Val.Int vy) -> Val.Int (vx + vy)
			       | _ -> raise (Error "Type checking error in Add."))
    | (Expr.Sub, [x;y])  -> (match (eval' x env, eval' y env) with
			       | (Val.Int vx, Val.Int vy) -> Val.Int (vx - vy)
			       | _ -> raise (Error "Type checking error in Sub."))
    | (Expr.Mul, [x;y])  -> (match (eval' x env, eval' y env) with
			       | (Val.Int vx, Val.Int vy) -> Val.Int (vx * vy)
			       | _ -> raise (Error "Type checking error in Mul."))
    | (Expr.Div, [x;y])  -> (match (eval' x env, eval' y env) with
			       | (Val.Int vx, Val.Int vy) -> Val.Int (vx / vy)
			       | _ -> raise (Error "Type checking error in Div."))
    | (Expr.Not, [x])    -> (match eval' x env with
			       | Val.Bool vx -> Val.Bool (not vx)
			       | _ -> raise (Error "Type checking error in Not."))
    | (Expr.And, [x;y])  -> (match (eval' x env, eval' y env) with
			       | (Val.Bool vx, Val.Bool vy) -> Val.Bool (vx && vy)
			       | _ -> raise (Error "Type checking error in And."))
    | (Expr.Eq, [x;y])   ->  Val.Bool ((eval' x env) = (eval' y env))
    | (Expr.Gt, [x;y])   -> (match (eval' x env, eval' y env) with
			       | (Val.Int vx, Val.Int vy) -> Val.Bool (vx > vy)
			       | _ -> raise (Error "Type checking error in Gt."))
    | (Expr.If, [c;x;y]) -> (match eval' c env with
			       | Val.Bool true  -> eval' x env
			       | Val.Bool false -> eval' y env
			       | _ -> raise (Error "Type checking error in If."))
    | _                  ->  raise (Error "Syntax error.")

  in eval' e (Env.create ())
